from royalnet.commands import *


class PingCommand(Command):
    name: str = "ping"

    description: str = "Get a pong response."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        await data.reply("🏓 Pong!")



import re
from royalnet.commands import *
from .dndnew import DndnewCommand
from ..tables import DndCharacter, DndActiveCharacter


class DndeditCommand(DndnewCommand):
    name: str = "dndedit"

    description: str = "Edit the active DnD character."

    aliases = ["de", "dnde", "edit", "dedit"]

    tables = {DndCharacter, DndActiveCharacter}

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        character_sheet = args.joined()

        if character_sheet == "":
            await data.reply(self._syntax())
            return

        author = await data.get_author(error_if_none=True)
        if author.dnd_active_character is None:
            raise CommandError("You don't have an active character.")

        char: DndCharacter = author.dnd_active_character.character

        arguments = self._parse(character_sheet)
        for key in arguments:
            char.__setattr__(key, arguments[key])

        await data.session_commit()

        await data.reply(f"✅ Edit successful!")



from sqlalchemy import Column, \
                       Integer, \
                       String, \
                       BigInteger, \
                       ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr
# noinspection PyUnresolvedReferences
from .users import User


class Discord:
    __tablename__ = "discord"

    @declared_attr
    def royal_id(self):
        return Column(Integer, ForeignKey("users.uid"))

    @declared_attr
    def discord_id(self):
        return Column(BigInteger, primary_key=True)

    @declared_attr
    def username(self):
        return Column(String)

    @declared_attr
    def discriminator(self):
        return Column(String)

    @declared_attr
    def avatar_hash(self):
        return Column(String)

    @declared_attr
    def royal(self):
        return relationship("User", backref="discord")

    def __repr__(self):
        return f"<Discord {str(self)}>"

    def __str__(self):
        return f"[c]discord:{self.full_username()}[/c]"

    def full_username(self):
        return f"{self.username}#{self.discriminator}"



# Imports go here!
from .ciaoruozi import CiaoruoziCommand
from .color import ColorCommand
from .cv import CvCommand
from .diario import DiarioCommand
from .rage import RageCommand
from .reminder import ReminderCommand
from .ship import ShipCommand
from .smecds import SmecdsCommand
from .videochannel import VideochannelCommand
from .trivia import TriviaCommand
from .matchmaking import MatchmakingCommand
from .pause import PauseCommand
from .play import PlayCommand
from .playmode import PlaymodeCommand
from .queue import QueueCommand
from .skip import SkipCommand
from .summon import SummonCommand
from .youtube import YoutubeCommand
from .soundcloud import SoundcloudCommand
from .zawarudo import ZawarudoCommand
from .emojify import EmojifyCommand
from .leagueoflegends import LeagueoflegendsCommand

# Enter the commands of your Pack here!
available_commands = [
    CiaoruoziCommand,
    ColorCommand,
    CvCommand,
    DiarioCommand,
    RageCommand,
    ReminderCommand,
    ShipCommand,
    SmecdsCommand,
    VideochannelCommand,
    TriviaCommand,
    MatchmakingCommand,
    PauseCommand,
    PlayCommand,
    PlaymodeCommand,
    QueueCommand,
    SkipCommand,
    SummonCommand,
    YoutubeCommand,
    SoundcloudCommand,
    ZawarudoCommand,
    EmojifyCommand,
    LeagueoflegendsCommand,
]

# Don't change this, it should automatically generate __all__
__all__ = [command.__class__.__qualname__ for command in available_commands]



from royalnet.commands import *
from royalnet.version import semantic


class VersionCommand(Command):
    name: str = "version"

    description: str = "Get the current Royalnet version."

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        message = f"ℹ️ Royalnet {semantic}\n"
        if "69" in message:
            message += "(Nice.)"
        await data.reply(message)



import discord
import sentry_sdk
import logging as _logging
import asyncio
from .generic import GenericBot
from ..utils import *
from ..error import *
from ..audio import *
from ..commands import *

log = _logging.getLogger(__name__)

# TODO: Load the opus library
if not discord.opus.is_loaded():
    log.error("Opus is not loaded. Weird behaviour might emerge.")


class MusicData:
    def __init__(self):
        self.playmode: playmodes.PlayMode = playmodes.Playlist()
        self.voice_client: typing.Optional[discord.VoiceClient] = None

    def queue_preview(self):
        return self.playmode.queue_preview()


class DiscordBot(GenericBot):
    """A bot that connects to `Discord <https://discordapp.com/>`_."""
    interface_name = "discord"

    def _init_voice(self):
        """Initialize the variables needed for the connection to voice chat."""
        log.debug(f"Creating music_data dict")
        self.music_data: typing.Dict[discord.Guild, MusicData] = {}

    def _interface_factory(self) -> typing.Type[CommandInterface]:
        # noinspection PyPep8Naming
        GenericInterface = super()._interface_factory()

        # noinspection PyMethodParameters,PyAbstractClass
        class DiscordInterface(GenericInterface):
            name = self.interface_name
            prefix = "!"

        return DiscordInterface

    def _data_factory(self) -> typing.Type[CommandData]:
        # noinspection PyMethodParameters,PyAbstractClass
        class DiscordData(CommandData):
            def __init__(data, interface: CommandInterface, message: discord.Message):
                super().__init__(interface)
                data.message = message

            async def reply(data, text: str):
                await data.message.channel.send(discord_escape(text))

            async def get_author(data, error_if_none=False):
                user: discord.Member = data.message.author
                query = data.session.query(self.master_table)
                for link in self.identity_chain:
                    query = query.join(link.mapper.class_)
                query = query.filter(self.identity_column == user.id)
                result = await asyncify(query.one_or_none)
                if result is None and error_if_none:
                    raise CommandError("You must be registered to use this command.")
                return result

            async def delete_invoking(data, error_if_unavailable=False):
                await data.message.delete()

        return DiscordData

    def _bot_factory(self) -> typing.Type[discord.Client]:
        """Create a custom DiscordClient class inheriting from :py:class:`discord.Client`."""
        log.debug(f"Creating DiscordClient")

        # noinspection PyMethodParameters
        class DiscordClient(discord.Client):
            async def vc_connect_or_move(cli, channel: discord.VoiceChannel):
                music_data = self.music_data.get(channel.guild)
                if music_data is None:
                    # Create a MusicData object
                    music_data = MusicData()
                    self.music_data[channel.guild] = music_data
                    # Connect to voice
                    log.debug(f"Connecting to Voice in {channel}")
                    try:
                        music_data.voice_client = await channel.connect(reconnect=False, timeout=10)
                    except Exception:
                        log.warning(f"Failed to connect to Voice in {channel}")
                        del self.music_data[channel.guild]
                        raise
                    else:
                        log.debug(f"Connected to Voice in {channel}")
                else:
                    if music_data.voice_client is None:
                        # TODO: change exception type
                        raise Exception("Another connection attempt is already in progress.")
                    # Try to move to a different channel
                    voice_client = music_data.voice_client
                    log.debug(f"Moving {voice_client} to {channel}")
                    await voice_client.move_to(channel)
                    log.debug(f"Moved {voice_client} to {channel}")

            async def on_message(cli, message: discord.Message):
                self.loop.create_task(cli._handle_message(message))

            async def _handle_message(cli, message: discord.Message):
                text = message.content
                # Skip non-text messages
                if not text:
                    return
                # Skip non-command updates
                if not text.startswith("!"):
                    return
                # Skip bot messages
                author: typing.Union[discord.User] = message.author
                if author.bot:
                    return
                # Find and clean parameters
                command_text, *parameters = text.split(" ")
                # Don't use a case-sensitive command name
                command_name = command_text.lower()
                # Find the command
                try:
                    command = self.commands[command_name]
                except KeyError:
                    # Skip the message
                    return
                # Prepare data
                data = self._Data(interface=command.interface, message=message)
                # Call the command
                log.debug(f"Calling command '{command.name}'")
                with message.channel.typing():
                    # Run the command
                    try:
                        await command.run(CommandArgs(parameters), data)
                    except InvalidInputError as e:
                        await data.reply(f":warning: {e.message}\n"
                                         f"Syntax: [c]/{command.name} {command.syntax}[/c]")
                    except UnsupportedError as e:
                        await data.reply(f":warning: {e.message}")
                    except CommandError as e:
                        await data.reply(f":warning: {e.message}")
                    except Exception as e:
                        sentry_sdk.capture_exception(e)
                        error_message = f"🦀 [b]{e.__class__.__name__}[/b] 🦀\n"
                        error_message += '\n'.join(e.args)
                        await data.reply(error_message)
                # Close the data session
                await data.session_close()

            async def on_connect(cli):
                log.debug("Connected to Discord")

            async def on_disconnect(cli):
                log.error("Disconnected from Discord!")

            async def on_ready(cli) -> None:
                log.debug("Connection successful, client is ready")
                await cli.change_presence(status=discord.Status.online)

            def find_guild_by_name(cli, name: str) -> typing.List[discord.Guild]:
                """Find the :py:class:`discord.Guild` with the specified name (case insensitive)."""
                all_guilds: typing.List[discord.Guild] = cli.guilds
                matching_channels: typing.List[discord.Guild] = []
                for guild in all_guilds:
                    if guild.name.lower() == name.lower():
                        matching_channels.append(guild)
                return matching_channels

            def find_channel_by_name(cli,
                                     name: str,
                                     guild: typing.Optional[discord.Guild] = None) -> typing.List[discord.abc.GuildChannel]:
                """Find the :py:class:`TextChannel`, :py:class:`VoiceChannel` or :py:class:`CategoryChannel` with the
                specified name (case insensitive).

                You can specify a guild to only find channels in that specific guild."""
                if guild is not None:
                    all_channels = guild.channels
                else:
                    all_channels: typing.List[discord.abc.GuildChannel] = cli.get_all_channels()
                matching_channels: typing.List[discord.abc.GuildChannel] = []
                for channel in all_channels:
                    if not (isinstance(channel, discord.TextChannel)
                            or isinstance(channel, discord.VoiceChannel)
                            or isinstance(channel, discord.CategoryChannel)):
                        continue
                    if channel.name.lower() == name.lower():
                        matching_channels.append(channel)
                return matching_channels

            def find_voice_client_by_guild(cli, guild: discord.Guild) -> typing.Optional[discord.VoiceClient]:
                """Find the :py:class:`discord.VoiceClient` belonging to a specific :py:class:`discord.Guild`."""
                for voice_client in cli.voice_clients:
                    if voice_client.guild == guild:
                        return voice_client
                return None

        return DiscordClient

    def _init_client(self):
        """Create an instance of the DiscordClient class created in :py:func:`royalnet.bots.DiscordBot._bot_factory`."""
        log.debug(f"Creating DiscordClient instance")
        self._Client = self._bot_factory()
        self.client = self._Client()

    def _initialize(self):
        super()._initialize()
        self._init_client()
        self._init_voice()

    async def run(self):
        """Login to Discord, then run the bot."""
        if not self.initialized:
            self._initialize()
        log.debug("Getting Discord secret")
        token = self.get_secret("discord")
        log.info(f"Logging in to Discord")
        await self.client.login(token)
        log.info(f"Connecting to Discord")
        await self.client.connect()

    async def add_to_music_data(self, dfiles: typing.List[YtdlDiscord], guild: discord.Guild):
        """Add a list of :py:class:`royalnet.audio.YtdlDiscord` to the corresponding music_data object."""
        guild_music_data = self.music_data[guild]
        if guild_music_data is None:
            raise CommandError(f"No music_data has been created for guild {guild}")
        for dfile in dfiles:
            log.debug(f"Adding {dfile} to music_data")
            await asyncify(dfile.ready_up)
            guild_music_data.playmode.add(dfile)
        if guild_music_data.playmode.now_playing is None:
            await self.advance_music_data(guild)

    async def advance_music_data(self, guild: discord.Guild):
        """Try to play the next song, while it exists. Otherwise, just return."""
        guild_music_data: MusicData = self.music_data[guild]
        voice_client: discord.VoiceClient = guild_music_data.voice_client
        next_source: discord.AudioSource = await guild_music_data.playmode.next()
        await self.update_activity_with_source_title()
        if next_source is None:
            log.debug(f"Ending playback chain")
            return

        def advance(error=None):
            if error:
                voice_client.disconnect(force=True)
                log.error(f"Error while advancing music_data: {error}")
                return
            self.loop.create_task(self.advance_music_data(guild))

        log.debug(f"Starting playback of {next_source}")
        voice_client.play(next_source, after=advance)

    async def update_activity_with_source_title(self):
        """Change the bot's presence (using :py:func:`discord.Client.change_presence`) to match the current listening status.

        If multiple guilds are using the bot, the bot will always have an empty presence."""
        if len(self.music_data) != 1:
            # Multiple guilds are using the bot, do not display anything
            log.debug(f"Updating current Activity: setting to None, as multiple guilds are using the bot")
            await self.client.change_presence(status=discord.Status.online)
            return
        play_mode: playmodes.PlayMode = self.music_data[list(self.music_data)[0]].playmode
        now_playing = play_mode.now_playing
        if now_playing is None:
            # No songs are playing now
            log.debug(f"Updating current Activity: setting to None, as nothing is currently being played")
            await self.client.change_presence(status=discord.Status.online)
            return
        log.debug(f"Updating current Activity: listening to {now_playing.info.title}")
        await self.client.change_presence(activity=discord.Activity(name=now_playing.info.title,
                                                                    type=discord.ActivityType.listening),
                                          status=discord.Status.online)



import typing
import pickle
import datetime
import discord
from royalnet.commands import *
from royalnet.utils import asyncify
from royalnet.audio import YtdlDiscord
from royalnet.bots import DiscordBot


class SoundcloudCommand(Command):
    name: str = "soundcloud"

    aliases = ["sc"]

    description: str = "Cerca una canzone su Soundcloud e la aggiunge alla coda della chat vocale."

    syntax = "[ [guild] ] {url}"

    @staticmethod
    async def _legacy_soundcloud_handler(bot: "DiscordBot", guild_name: typing.Optional[str], search: str):
        # Find the matching guild
        if guild_name:
            guilds: typing.List[discord.Guild] = bot.client.find_guild_by_name(guild_name)
        else:
            guilds = bot.client.guilds
        if len(guilds) == 0:
            raise CommandError("Server non trovato.")
        if len(guilds) > 1:
            raise CommandError("Il nome del server è ambiguo.")
        guild = list(bot.client.guilds)[0]
        # Ensure the guild has a PlayMode before adding the file to it
        if not bot.music_data.get(guild):
            raise CommandError("Il bot non è in nessun canale vocale.")
        # Create url
        ytdl_args = {
            "format": "bestaudio/best",
            "outtmpl": f"./downloads/{datetime.datetime.now().timestamp()}_%(title)s.%(ext)s"
        }
        # Start downloading
        dfiles: typing.List[YtdlDiscord] = await asyncify(YtdlDiscord.create_from_url, f'scsearch:{search}',
                                                          **ytdl_args)
        await bot.add_to_music_data(dfiles, guild)
        # Create response dictionary
        return {
            "videos": [{
                "title": dfile.info.title,
                "discord_embed_pickle": str(pickle.dumps(dfile.info.to_discord_embed()))
            } for dfile in dfiles]
        }

    _event_name = "_legacy_soundcloud"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if interface.name == "discord":
            interface.register_herald_action(self._event_name, self._legacy_soundcloud_handler)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, search = args.match(r"(?:\[(.+)])?\s*<?(.+)>?")
        if search.startswith("http://") or search.startswith("https://"):
            raise CommandError(f"Il comando [c]{self.interface.prefix}soundcloud[/c] funziona solo per cercare audio su"
                               f" Soundcloud con un dato nome.\n"
                               f"Se vuoi riprodurre una canzone da un URL, usa [c]{self.interface.prefix}play[/c]!")
        response = await self.interface.call_herald_action("discord", self._event_name, {
                                                               "guild_name": guild_name,
                                                               "search": search
                                                           })
        if len(response["videos"]) == 0:
            raise CommandError(f"Il video non può essere scaricato a causa di un blocco imposto da Soundcloud.")
        for video in response["videos"]:
            if self.interface.name == "discord":
                # This is one of the unsafest things ever
                embed = pickle.loads(eval(video["discord_embed_pickle"]))
                await data.message.channel.send(content="▶️ Aggiunto alla coda:", embed=embed)
            else:
                await data.reply(f"▶️ Aggiunto alla coda: [i]{video['title']}[/i]")



import typing
import discord
from royalnet.commands import *


class VideochannelCommand(Command):
    name: str = "videochannel"

    aliases = ["golive", "live", "video"]

    description: str = "Converti il canale vocale in un canale video."

    syntax = "[channelname]"

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name == "discord":
            bot: discord.Client = self.interface.bot
            message: discord.Message = data.message
            channel_name: str = args.optional(0)
            if channel_name:
                guild: typing.Optional[discord.Guild] = message.guild
                if guild is not None:
                    channels: typing.List[discord.abc.GuildChannel] = guild.channels
                else:
                    channels = bot.get_all_channels()
                matching_channels: typing.List[discord.VoiceChannel] = []
                for channel in channels:
                    if isinstance(channel, discord.VoiceChannel):
                        if channel.name == channel_name:
                            matching_channels.append(channel)
                if len(matching_channels) == 0:
                    raise CommandError("Non esiste alcun canale vocale con il nome specificato.")
                elif len(matching_channels) > 1:
                    raise CommandError("Esiste più di un canale vocale con il nome specificato.")
                channel = matching_channels[0]
            else:
                author: discord.Member = message.author
                voice: typing.Optional[discord.VoiceState] = author.voice
                if voice is None:
                    raise CommandError("Non sei connesso a nessun canale vocale.")
                channel = voice.channel
                if author.is_on_mobile():
                    await data.reply(f"📹 Per entrare in modalità video, clicca qui:\n"
                                     f"<https://discordapp.com/channels/{channel.guild.id}/{channel.id}>\n"
                                     f"[b]Attenzione: la modalità video non funziona su Android e iOS![/b]")
                    return
            await data.reply(f"📹 Per entrare in modalità video, clicca qui:\n"
                             f"<https://discordapp.com/channels/{channel.guild.id}/{channel.id}>")
        else:
            raise UnsupportedError()



"""A User Games Diario viewer :py:class:`royalnet.web.Royalprint`."""

import flask as f
import os
from ...royalprint import Royalprint
from ...shortcuts import error
from royalnet.packs.common.tables import User
from royalnet.packs.royal.tables import Diario


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("diarioview", __name__, url_prefix="/diario", template_folder=tmpl_dir,
                required_tables={User, Diario})


@rp.route("/", defaults={"page": 1})
@rp.route("/<int:page>")
def diarioview_page(page):
    alchemy, alchemy_session = f.current_app.config["ALCHEMY"], f.current_app.config["ALCHEMY_SESSION"]
    if page < 1:
        return error(404, "Il numero di pagina deve essere maggiore di 0.")
    entries = alchemy_session.query(alchemy.Diario).order_by(alchemy.Diario.diario_id.desc()).offset((page - 1) * 1000).limit(1000).all()
    if len(entries) == 0:
        return error(404, "Non ci sono righe di diario in questa pagina (e in tutte le successive).")
    return f.render_template("diarioview_page.html", page=page, entries=entries)



from royalnet.commands import *
from royalnet.utils import asyncify
from ..tables import DndCharacter, DndActiveCharacter


class DndinfoCommand(Command):
    name: str = "dndinfo"

    description: str = "Display the character sheet of the active DnD character."

    aliases = ["di", "dndi", "info", "dinfo"]

    tables = {DndCharacter, DndActiveCharacter}

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        author = await data.get_author(error_if_none=True)
        if author.dnd_active_character is None:
            raise CommandError("You don't have an active character.")
        await data.reply(author.dnd_active_character.character.character_sheet())



"""Various bot interfaces, and a common class to create new ones."""

from .generic import GenericBot
from .telegram import TelegramBot
from .discord import DiscordBot

__all__ = ["TelegramBot", "DiscordBot", "GenericBot"]



import dice
from royalnet.commands import *


class DiceCommand(Command):
    name: str = "dice"

    description: str = "Roll a dice, using 'dice'."

    syntax = "{dice}"

    aliases = ["d"]

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        dice_str = args.joined(require_at_least=1)
        try:
            roll = dice.roll(dice_str)
        except dice.DiceFatalException as e:
            raise CommandError(e.msg)
        except dice.DiceException as e:
            raise CommandError(e.msg)
        except dice.DiceBaseException as e:
            raise CommandError(str(e))
        try:
            result = list(roll)
        except TypeError:
            result = [roll]
        message = f"🎲 {dice_str}"
        total = 0
        if len(result) > 1:
            message += f" = "
            for index, die in enumerate(result):
                message += f"{die}"
                total += int(die)
                if (index + 1) < len(result):
                    message += "+"
        else:
            total += int(result[0])
        message += f" = [b]{total}[/b]"
        await data.reply(message)



import typing
import pickle
import discord
from royalnet.commands import *
from royalnet.utils import numberemojiformat
from royalnet.bots import DiscordBot


class QueueCommand(Command):
    name: str = "queue"

    aliases = ["q"]

    description: str = "Visualizza la coda di riproduzione attuale."

    syntax = "[ [guild] ]"

    @staticmethod
    async def _legacy_queue_handler(bot: "DiscordBot", guild_name: typing.Optional[str]):
        # Find the matching guild
        if guild_name:
            guilds: typing.List[discord.Guild] = bot.client.find_guild_by_name(guild_name)
        else:
            guilds = bot.client.guilds
        if len(guilds) == 0:
            raise CommandError("No guilds with the specified name found.")
        if len(guilds) > 1:
            raise CommandError("Multiple guilds with the specified name found.")
        guild = list(bot.client.guilds)[0]
        # Check if the guild has a PlayMode
        playmode = bot.music_data.get(guild)
        if not playmode:
            return {
                "type": None
            }
        try:
            queue = playmode.queue_preview()
        except NotImplementedError:
            return {
                "type": playmode.__class__.__name__
            }
        return {
            "type": playmode.__class__.__name__,
            "queue":
                {
                    "strings": [str(dfile.info) for dfile in queue],
                    "pickled_embeds": str(pickle.dumps([dfile.info.to_discord_embed() for dfile in queue]))
                }
        }

    _event_name = "_legacy_queue"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if interface.name == "discord":
            interface.register_herald_action(self._event_name, self._legacy_queue_handler)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, = args.match(r"(?:\[(.+)])?")
        response = await self.interface.call_herald_action("discord", self._event_name, {"guild_name": guild_name})
        if response["type"] is None:
            await data.reply("ℹ️ Non c'è nessuna coda di riproduzione attiva al momento.")
            return
        elif "queue" not in response:
            await data.reply(f"ℹ️ La coda di riproduzione attuale ([c]{response['type']}[/c]) non permette l'anteprima.")
            return
        if response["type"] == "Playlist":
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Questa [c]Playlist[/c] è vuota."
            else:
                message = f"ℹ️ Questa [c]Playlist[/c] contiene {len(response['queue']['strings'])} elementi, e i prossimi saranno:\n"
        elif response["type"] == "Pool":
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Questo [c]Pool[/c] è vuoto."
            else:
                message = f"ℹ️ Questo [c]Pool[/c] contiene {len(response['queue']['strings'])} elementi, tra cui:\n"
        elif response["type"] == "Layers":
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Nessun elemento è attualmente in riproduzione, pertanto non ci sono [c]Layers[/c]:"
            else:
                message = f"ℹ️ I [c]Layers[/c] dell'elemento attualmente in riproduzione sono {len(response['queue']['strings'])}, tra cui:\n"
        else:
            if len(response["queue"]["strings"]) == 0:
                message = f"ℹ️ Il PlayMode attuale, [c]{response['type']}[/c], è vuoto.\n"
            else:
                message = f"ℹ️ Il PlayMode attuale, [c]{response['type']}[/c], contiene {len(response['queue']['strings'])} elementi:\n"
        if self.interface.name == "discord":
            await data.reply(message)
            for embed in pickle.loads(eval(response["queue"]["pickled_embeds"]))[:5]:
                await data.message.channel.send(embed=embed)
        else:
            message += numberemojiformat(response["queue"]["strings"][:10])
            await data.reply(message)



# Imports go here!
from .dndactivecharacters import DndActiveCharacter
from .dndcharacters import DndCharacter

# Enter the tables of your Pack here!
available_tables = [
    DndActiveCharacter,
    DndCharacter,
]

# Don't change this, it should automatically generate __all__
__all__ = [table.__class__.__qualname__ for table in available_tables]



import click
import typing
import importlib
import royalnet as r
import royalherald as rh
import multiprocessing
import keyring


@click.command()
@click.option("--telegram/--no-telegram", default=None,
              help="Enable/disable the Telegram module.")
@click.option("--discord/--no-discord", default=None,
              help="Enable/disable the Discord module.")
@click.option("-d", "--database", type=str, default=None,
              help="The PostgreSQL database path.")
@click.option("-p", "--packs", type=str, multiple=True, default=[],
              help="The names of the Packs that should be used.")
@click.option("-n", "--network-address", type=str, default=None,
              help="The Network server URL to connect to.")
@click.option("-l", "--local-network-server", is_flag=True, default=False,
              help="Locally run a Network server and bind it to port 44444. Overrides -n.")
@click.option("-s", "--secrets-name", type=str, default="__default__",
              help="The name in the keyring that the secrets are stored with.")
@click.option("-v", "--verbose", is_flag=True, default=False,
              help="Print all possible debug information.")
def run(telegram: typing.Optional[bool],
        discord: typing.Optional[bool],
        database: typing.Optional[str],
        packs: typing.Tuple[str],
        network_address: typing.Optional[str],
        local_network_server: bool,
        secrets_name: str,
        verbose: bool):

    # Get the network password
    network_password = keyring.get_password(f"Royalnet/{secrets_name}", "network")

    # Get the sentry dsn
    sentry_dsn = keyring.get_password(f"Royalnet/{secrets_name}", "sentry")

    # Enable / Disable interfaces
    interfaces = {
        "telegram": telegram,
        "discord": discord
    }
    # If any interface is True, then the undefined ones should be False
    if any(interfaces[name] is True for name in interfaces):
        for name in interfaces:
            if interfaces[name] is None:
                interfaces[name] = False
    # Likewise, if any interface is False, then the undefined ones should be True
    elif any(interfaces[name] is False for name in interfaces):
        for name in interfaces:
            if interfaces[name] is None:
                interfaces[name] = True
    # Otherwise, if no interfaces are specified, all should be enabled
    else:
        assert all(interfaces[name] is None for name in interfaces)
        for name in interfaces:
            interfaces[name] = True

    server_process: typing.Optional[multiprocessing.Process] = None
    # Start the network server
    if local_network_server:
        server_process = multiprocessing.Process(name="Network Server",
                                                 target=rh.Server("0.0.0.0", 44444, network_password).run_blocking,
                                                 daemon=True)
        server_process.start()
        network_address = "ws://127.0.0.1:44444/"

    # Create a Royalnet configuration
    network_config: typing.Optional[rh.Config] = None
    if network_address is not None:
        network_config = rh.Config(network_address, network_password)

    # Create a Alchemy configuration
    telegram_db_config: typing.Optional[r.database.DatabaseConfig] = None
    discord_db_config: typing.Optional[r.database.DatabaseConfig] = None
    if database is not None:
        telegram_db_config = r.database.DatabaseConfig(database,
                                                       r.packs.common.tables.User,
                                                       r.packs.common.tables.Telegram,
                                                       "tg_id")
        discord_db_config = r.database.DatabaseConfig(database,
                                                      r.packs.common.tables.User,
                                                      r.packs.common.tables.Discord,
                                                      "discord_id")

    # Import command packs
    packs: typing.List[str] = list(packs)
    packs.append("royalnet.packs.common")  # common pack is always imported
    enabled_commands = []
    for pack in packs:
        imported = importlib.import_module(pack)
        try:
            imported_commands = imported.available_commands
        except AttributeError:
            raise click.ClickException(f"{pack} isn't a Royalnet Pack.")
        enabled_commands = [*enabled_commands, *imported_commands]

    telegram_process: typing.Optional[multiprocessing.Process] = None
    if interfaces["telegram"]:
        click.echo("\n@BotFather Commands String")
        for command in enabled_commands:
            click.echo(f"{command.name} - {command.description}")
        click.echo("")
        telegram_bot = r.bots.TelegramBot(network_config=network_config,
                                          database_config=telegram_db_config,
                                          sentry_dsn=sentry_dsn,
                                          commands=enabled_commands,
                                          secrets_name=secrets_name)
        telegram_process = multiprocessing.Process(name="Telegram Interface",
                                                   target=telegram_bot.run_blocking,
                                                   args=(verbose,),
                                                   daemon=True)
        telegram_process.start()

    discord_process: typing.Optional[multiprocessing.Process] = None
    if interfaces["discord"]:
        discord_bot = r.bots.DiscordBot(network_config=network_config,
                                        database_config=discord_db_config,
                                        sentry_dsn=sentry_dsn,
                                        commands=enabled_commands,
                                        secrets_name=secrets_name)
        discord_process = multiprocessing.Process(name="Discord Interface",
                                                  target=discord_bot.run_blocking,
                                                  args=(verbose,),
                                                  daemon=True)
        discord_process.start()

    click.echo("Royalnet processes have been started. You can force-quit by pressing Ctrl+C.")
    if server_process is not None:
        server_process.join()
    if telegram_process is not None:
        telegram_process.join()
    if discord_process is not None:
        discord_process.join()


if __name__ == "__main__":
    run()



# This is a template Pack __init__. You can use this without changing anything in other packages too!

from .commands import available_commands
from .tables import available_tables

__all__ = ["commands", "tables", "available_commands", "available_tables"]



# Imports go here!
from .users import User
from .telegram import Telegram
from .discord import Discord

# Enter the tables of your Pack here!
available_tables = [
    User,
    Telegram,
    Discord
]

# Don't change this, it should automatically generate __all__
__all__ = [table.__class__.__qualname__ for table in available_tables]



import re
# noinspection PyUnresolvedReferences
from royalnet.commands import *
from ..tables import DndCharacter
from ..utils import DndProficiencyType


class DndnewCommand(Command):
    name: str = "dndnew"

    description: str = "Create a new DnD character."

    aliases = ["dn", "dndn", "new", "dnew"]

    syntax = "{name}\n{character_sheet}"

    tables = {DndCharacter}

    def _search_value(self, name: str, string: str):
        return re.search(r"\s*" + name + r"\s*([0-9]+)\s*", string, re.IGNORECASE)

    def _parse(self, character_sheet: str) -> dict:
        columns = list(self.alchemy.DndCharacter.__table__.columns)
        column_names = [column.name for column in columns if (not column.primary_key and
                                                              not column.foreign_keys and
                                                              column.name != "name")]
        arguments = {}
        for column_name in column_names:
            match = self._search_value(column_name, character_sheet)
            if match:
                if column_name.endswith("_proficiency"):
                    arguments[column_name] = DndProficiencyType(float(match.group(1)))
                else:
                    arguments[column_name] = match.group(1)
        return arguments

    def _syntax(self) -> str:
        columns = list(self.alchemy.DndCharacter.__table__.columns)
        column_names = [column.name for column in columns if (not column.primary_key and
                                                              not column.foreign_keys and
                                                              column.name != "name")]
        message = "ℹ️ How to create a new character:\n[p]/dndnew YOUR_CHARACTER_NAME\n"
        for column_name in column_names:
            message += f"{column_name} _\n"
        message += "[/p]"
        return message

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        character_sheet = args.joined()

        if character_sheet == "":
            await data.reply(self._syntax())
            return

        creator = await data.get_author()

        name, rest = character_sheet.split("\n", 1)

        character = self.alchemy.DndCharacter(name=name, creator=creator, **self._parse(rest))
        data.session.add(character)

        try:
            await data.session_commit()
        except Exception as err:
            # THIS IS INTENDED
            if err.__class__.__name__ == "IntegrityError":
                param_name = re.search(r'in column "(\S+)"', err.args[0]).group(1)
                raise CommandError(f"Mandatory parameter '{param_name}' is missing.")
            raise

        await data.reply(f"✅ Character [b]{character.name}[/b] created!")



import typing
import discord
from royalnet.commands import *
from royalnet.bots import DiscordBot


class SummonCommand(Command):
    name: str = "summon"

    aliases = ["cv"]

    description: str = "Evoca il bot in un canale vocale."

    syntax: str = "[nomecanale]"

    @staticmethod
    async def _legacy_summon_handler(bot: "DiscordBot", channel_name: str):
        """Handle a summon Royalnet request.
         That is, join a voice channel, or move to a different one if that is not possible."""
        channels = bot.client.find_channel_by_name(channel_name)
        if len(channels) < 1:
            raise CommandError(f"Nessun canale vocale con il nome [c]{channel_name}[/c] trovato.")
        channel = channels[0]
        if not isinstance(channel, discord.VoiceChannel):
            raise CommandError(f"Il canale [c]{channel}[/c] non è un canale vocale.")
        bot.loop.create_task(bot.client.vc_connect_or_move(channel))
        return {}

    _event_name = "_legacy_summon"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if interface.name == "discord":
            interface.register_herald_action(self._event_name, self._legacy_summon_handler)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        if self.interface.name == "discord":
            bot = self.interface.bot.client
            message: discord.Message = data.message
            channel_name: str = args.optional(0)
            if channel_name:
                guild: typing.Optional[discord.Guild] = message.guild
                if guild is not None:
                    channels: typing.List[discord.abc.GuildChannel] = guild.channels
                else:
                    channels = bot.get_all_channels()
                matching_channels: typing.List[discord.VoiceChannel] = []
                for channel in channels:
                    if isinstance(channel, discord.VoiceChannel):
                        if channel.name == channel_name:
                            matching_channels.append(channel)
                if len(matching_channels) == 0:
                    await data.reply("⚠️ Non esiste alcun canale vocale con il nome specificato.")
                    return
                elif len(matching_channels) > 1:
                    await data.reply("⚠️ Esiste più di un canale vocale con il nome specificato.")
                    return
                channel = matching_channels[0]
            else:
                author: discord.Member = message.author
                try:
                    voice: typing.Optional[discord.VoiceState] = author.voice
                except AttributeError:
                    await data.reply("⚠️ Non puoi evocare il bot da una chat privata!")
                    return
                if voice is None:
                    await data.reply("⚠️ Non sei connesso a nessun canale vocale!")
                    return
                channel = voice.channel
            await bot.vc_connect_or_move(channel)
            await data.reply(f"✅ Mi sono connesso in [c]#{channel.name}[/c].")
        else:
            channel_name: str = args[0].lstrip("#")
            response = await self.interface.call_herald_action("discord", self._event_name, {
                                                                   "channel_name": channel_name
                                                               })
            await data.reply(f"✅ Mi sono connesso in [c]#{channel_name}[/c].")



from royalnet.commands import *
from royalnet.utils import asyncify
from ..tables import DndCharacter, DndActiveCharacter


class DndactiveCommand(Command):
    name: str = "dndactive"

    description: str = "Set a DnD character as active."

    aliases = ["da", "dnda", "active", "dactive"]

    syntax = "{name|id}"

    tables = {DndCharacter, DndActiveCharacter}

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        identifier = args.optional(0)
        author = await data.get_author(error_if_none=True)
        if identifier is None:
            # Display the active character
            if author.dnd_active_character is None:
                await data.reply("ℹ️ You have no active characters.")
            else:
                await data.reply(f"ℹ️ You currently active character is [b]{author.dnd_active_character}[/b].")
            return
        try:
            identifier = int(identifier)
        except ValueError:
            # Find the character by name
            chars = await asyncify(data.session.query(self.alchemy.DndCharacter).filter_by(name=identifier).all)
            if len(chars) >= 2:
                char_string = "\n".join([f"[c]{char.character_id}[/c] (LV {char.level}) by {char.creator})" for char in chars])
                raise CommandError(f"Multiple characters share the name {identifier}, "
                                   f"please activate them using their id:\n{char_string}")
            elif len(chars) == 1:
                char = chars[0]
            else:
                char = None
        else:
            # Find the character by id
            char = await asyncify(data.session.query(self.alchemy.DndCharacter)
                                              .filter_by(character_id=identifier)
                                              .one_or_none)
        if char is None:
            raise CommandError("No character found.")
        # Check if the player already has an active character
        if author.dnd_active_character is None:
            # Create a new active character
            achar = self.alchemy.DndActiveCharacter(character=char, user=author)
            data.session.add(achar)
        else:
            # Change the active character
            author.dnd_active_character.character = char
        await data.session_commit()
        await data.reply(f"✅ Active character set to [b]{char}[/b]!")



import typing
import discord
from royalnet.commands import *
from royalnet.bots import DiscordBot


class SkipCommand(Command):
    name: str = "skip"

    aliases = ["s", "next", "n"]

    description: str = "Salta la canzone attualmente in riproduzione in chat vocale."

    syntax: str = "[ [guild] ]"

    @staticmethod
    async def _legacy_skip_handler(bot: "DiscordBot", guild_name: typing.Optional[str]):
        # Find the matching guild
        if guild_name:
            guilds: typing.List[discord.Guild] = bot.client.find_guild_by_name(guild_name)
        else:
            guilds = bot.client.guilds
        if len(guilds) == 0:
            raise CommandError("No guilds with the specified name found.")
        if len(guilds) > 1:
            raise CommandError("Multiple guilds with the specified name found.")
        guild = list(bot.client.guilds)[0]
        # Set the currently playing source as ended
        voice_client: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        if not (voice_client.is_playing() or voice_client.is_paused()):
            raise CommandError("Nothing to skip")
        # noinspection PyProtectedMember
        voice_client._player.stop()
        return {}

    _event_name = "_legacy_skip"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if interface.name == "discord":
            interface.register_herald_action(self._event_name, self._legacy_skip_handler)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, = args.match(r"(?:\[(.+)])?")
        await self.interface.call_herald_action("discord", self._event_name, {
                                                    "guild_name": guild_name
                                                })
        await data.reply(f"⏩ Richiesto lo skip della canzone attuale.")



from .commandinterface import CommandInterface
from .command import Command
from .commanddata import CommandData
from .commandargs import CommandArgs
from .commanderrors import CommandError, InvalidInputError, UnsupportedError, KeyboardExpiredError

__all__ = [
    "CommandInterface",
    "Command",
    "CommandData",
    "CommandArgs",
    "CommandError",
    "InvalidInputError",
    "UnsupportedError",
    "KeyboardExpiredError",
]



from .mmchoice import MMChoice
from .mminterfacedata import MMInterfaceData, MMInterfaceDataTelegram
from .leaguetier import LeagueTier
from .leaguerank import LeagueRank
from .leagueleague import LeagueLeague

__all__ = ["MMChoice", "MMInterfaceData", "MMInterfaceDataTelegram", "LeagueTier", "LeagueRank", "LeagueLeague"]



from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import *


class DndActiveCharacter:
    __tablename__ = "dndactivecharacters"

    @declared_attr
    def character_id(self):
        return Column(Integer, ForeignKey("dndcharacters.character_id"), primary_key=True)

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"), primary_key=True)

    @declared_attr
    def character(self):
        return relationship("DndCharacter", foreign_keys=self.character_id, backref="activated_by")

    @declared_attr
    def user(self):
        return relationship("User", foreign_keys=self.user_id, backref=backref("dnd_active_character", uselist=False))

    def __repr__(self):
        return f"<{self.__class__.__qualname__} for {self.user_id}: {self.character_id}>"



import click
import keyring


@click.command()
def run():
    click.echo("Welcome to the Royalnet configuration creator!")
    secrets_name = click.prompt("Desired secrets name", default="__default__")
    network = click.prompt("Network password", default="")
    if network:
        keyring.set_password(f"Royalnet/{secrets_name}", "network", network)
    telegram = click.prompt("Telegram Bot API token", default="")
    if telegram:
        keyring.set_password(f"Royalnet/{secrets_name}", "telegram", telegram)
    discord = click.prompt("Discord Bot API token", default="")
    if discord:
        keyring.set_password(f"Royalnet/{secrets_name}", "discord", discord)
    imgur = click.prompt("Imgur API token", default="")
    if imgur:
        keyring.set_password(f"Royalnet/{secrets_name}", "imgur", imgur)
    sentry = click.prompt("Sentry DSN", default="")
    if sentry:
        keyring.set_password(f"Royalnet/{secrets_name}", "sentry", sentry)
    leagueoflegends = click.prompt("League of Legends API Token", default="")
    if leagueoflegends:
        keyring.set_password(f"Royalnet/{secrets_name}", "leagueoflegends", leagueoflegends)


if __name__ == "__main__":
    run()



"""Minecraft server status :py:class:`royalnet.web.Royalprint`."""
import os
import flask as f
import socket
from ...royalprint import Royalprint
from ...shortcuts import error
from mcstatus import MinecraftServer


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("mcstatus", __name__, url_prefix="/mcstatus", template_folder=tmpl_dir)


@rp.route("/<server_str>")
def mcstatus_index(server_str: str):
    try:
        if ":" not in server_str:
            server_str += ":25565"
        server = MinecraftServer.lookup(server_str)
        status = server.status()
        try:
            query = server.query()
        except (socket.timeout, ConnectionRefusedError, OSError):
            query = None
    except socket.gaierror:
        return error(400, "L'indirizzo richiesto non è valido.")
    except (socket.timeout, ConnectionRefusedError, OSError):
        status = None
        query = None
    return f.render_template("mcstatus.html", server_str=server_str, status=status, query=query)



from sqlalchemy import Column, \
                       Integer, \
                       ForeignKey
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr


class TriviaScore:
    __tablename__ = "triviascores"

    @declared_attr
    def royal_id(self):
        return Column(Integer, ForeignKey("users.uid"), primary_key=True)

    @declared_attr
    def royal(self):
        return relationship("User", backref=backref("trivia_score", uselist=False))

    @declared_attr
    def correct_answers(self):
        return Column(Integer, nullable=False, default=0)

    @declared_attr
    def wrong_answers(self):
        return Column(Integer, nullable=False, default=0)

    @property
    def total_answers(self):
        return self.correct_answers + self.wrong_answers

    @property
    def offset(self):
        return self.correct_answers - self.wrong_answers

    @property
    def correct_rate(self):
        return self.correct_answers / self.total_answers

    def __repr__(self):
        return f"<TriviaScore of {self.royal}: ({self.correct_answers}|{self.wrong_answers})>"



import discord
import typing
from royalnet.commands import *
from royalnet.utils import andformat
from royalnet.bots import DiscordBot


class CvCommand(Command):
    name: str = "cv"

    description: str = "Elenca le persone attualmente connesse alla chat vocale."

    syntax: str = "[guildname] [all]"

    @staticmethod
    async def _legacy_cv_handler(bot: DiscordBot, guild_name: typing.Optional[str], everyone: bool):
        # Find the matching guild
        if guild_name:
            guilds: typing.List[discord.Guild] = bot.client.find_guild_by_name(guild_name)
        else:
            guilds = bot.client.guilds
        if len(guilds) == 0:
            raise CommandError("No guilds with the specified name found.")
        if len(guilds) > 1:
            raise CommandError("Multiple guilds with the specified name found.")
        guild = list(bot.client.guilds)[0]
        # Edit the message, sorted by channel
        discord_members = list(guild.members)
        channels = {0: None}
        members_in_channels = {0: []}
        message = ""
        # Find all the channels
        for member in discord_members:
            if member.voice is not None:
                channel = members_in_channels.get(member.voice.channel.id)
                if channel is None:
                    members_in_channels[member.voice.channel.id] = list()
                    channel = members_in_channels[member.voice.channel.id]
                    channels[member.voice.channel.id] = member.voice.channel
                channel.append(member)
            else:
                members_in_channels[0].append(member)
        # Edit the message, sorted by channel
        for channel in sorted(channels, key=lambda c: -c):
            members_in_channels[channel].sort(key=lambda x: x.nick if x.nick is not None else x.name)
            if channel == 0 and len(members_in_channels[0]) > 0:
                message += "[b]Non in chat vocale:[/b]\n"
            else:
                message += f"[b]In #{channels[channel].name}:[/b]\n"
            for member in members_in_channels[channel]:
                member: typing.Union[discord.User, discord.Member]
                # Ignore not-connected non-notable members
                if not everyone and channel == 0 and len(member.roles) < 2:
                    continue
                # Ignore offline members
                if member.status == discord.Status.offline and member.voice is None:
                    continue
                # Online status emoji
                if member.bot:
                    message += "🤖 "
                elif member.status == discord.Status.online:
                    message += "🔵 "
                elif member.status == discord.Status.idle:
                    message += "⚫ "
                elif member.status == discord.Status.dnd:
                    message += "🔴 "
                elif member.status == discord.Status.offline:
                    message += "⚪ "
                # Voice
                if channel != 0:
                    # Voice status
                    if member.voice.afk:
                        message += "💤 "
                    elif member.voice.self_deaf or member.voice.deaf:
                        message += "🔇 "
                    elif member.voice.self_mute or member.voice.mute:
                        message += "🔈 "
                    elif member.voice.self_video or member.voice.self_stream:
                        message += "🖥 "
                    else:
                        message += "🔊 "
                # Nickname
                # if member.nick is not None:
                #     message += f"[i]{member.nick}[/i]"
                # else:
                message += member.name
                # Game or stream
                if member.activity is not None:
                    if member.activity.type == discord.ActivityType.playing:
                        message += f" | 🎮 {member.activity.name}"
                        # Rich presence
                        try:
                            if member.activity.state is not None:
                                message += f" ({member.activity.state}" \
                                           f" | {member.activity.details})"
                        except AttributeError:
                            pass
                    elif member.activity.type == discord.ActivityType.streaming:
                        message += f" | 📡 {member.activity.url}"
                    elif member.activity.type == discord.ActivityType.listening:
                        if isinstance(member.activity, discord.Spotify):
                            if member.activity.title == member.activity.album:
                                message += f" | 🎧 {member.activity.title} ({andformat(member.activity.artists, final=' e ')})"
                            else:
                                message += f" | 🎧 {member.activity.title} ({member.activity.album} | {andformat(member.activity.artists, final=' e ')})"
                        else:
                            message += f" | 🎧 {member.activity.name}"
                    elif member.activity.type == discord.ActivityType.watching:
                        message += f" | 📺 {member.activity.name}"
                    else:
                        message += f" | ❓ Unknown activity"
                message += "\n"
            message += "\n"
        return {"response": message}

    _event_name = "_legacy_cv"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if interface.name == "discord":
            interface.register_herald_action(self._event_name, self._legacy_cv_handler)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        # noinspection PyTypeChecker
        guild_name, everyone = args.match(r"(?:\[(.+)])?\s*(\S+)?\s*")
        response = await self.interface.call_herald_action("discord", self._event_name, {
                                                               "guild_name": guild_name,
                                                               "everyone": everyone
                                                           })
        await data.reply(response["response"])



"""Homepage :py:class:`royalnet.web.Royalprint` of the User Games website."""
import flask as f
import os
from ...royalprint import Royalprint


tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates')
rp = Royalprint("home", __name__, template_folder="templates")


@rp.route("/")
def home_index():
    return f.render_template("home.html")



import sys
import asyncio
import logging
import sentry_sdk
import keyring
import royalnet.version
import royalherald as rh
from sentry_sdk.integrations.sqlalchemy import SqlalchemyIntegration
from sentry_sdk.integrations.aiohttp import AioHttpIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from ..utils import *
from ..database import *
from ..commands import *
from ..error import *


log = logging.getLogger(__name__)


class GenericBot:
    """A common bot class, to be used as base for the other more specific classes, such as
    :py:class:`royalnet.bots.TelegramBot` and :py:class:`royalnet.bots.DiscordBot`. """
    interface_name = NotImplemented

    def _init_commands(self) -> None:
        """Generate the ``packs`` dictionary required to handle incoming messages, and the ``network_handlers``
        dictionary required to handle incoming requests. """
        log.info(f"Registering packs...")
        self._Interface = self._interface_factory()
        self._Data = self._data_factory()
        self.commands = {}
        self.network_handlers: typing.Dict[str, typing.Callable[["GenericBot", typing.Any],
                                                                typing.Awaitable[typing.Optional[typing.Dict]]]] = {}
        for SelectedCommand in self.uninitialized_commands:
            interface = self._Interface()
            try:
                command = SelectedCommand(interface)
            except Exception as e:
                log.error(f"{e.__class__.__qualname__} during the registration of {SelectedCommand.__qualname__}")
                continue
            # Linking the command to the interface
            interface.command = command
            # Override the main command name, but warn if it's overriding something
            if f"{interface.prefix}{SelectedCommand.name}" in self.commands:
                log.warning(f"Overriding (already defined): {SelectedCommand.__qualname__} -> {interface.prefix}{SelectedCommand.name}")
            else:
                log.debug(f"Registering: {SelectedCommand.__qualname__} -> {interface.prefix}{SelectedCommand.name}")
            self.commands[f"{interface.prefix}{SelectedCommand.name}"] = command
            # Register aliases, but don't override anything
            for alias in SelectedCommand.aliases:
                if f"{interface.prefix}{alias}" not in self.commands:
                    log.debug(f"Aliasing: {SelectedCommand.__qualname__} -> {interface.prefix}{alias}")
                    self.commands[f"{interface.prefix}{alias}"] = self.commands[f"{interface.prefix}{SelectedCommand.name}"]
                else:
                    log.info(f"Ignoring (already defined): {SelectedCommand.__qualname__} -> {interface.prefix}{alias}")

    def _interface_factory(self) -> typing.Type[CommandInterface]:
        # noinspection PyAbstractClass,PyMethodParameters
        class GenericInterface(CommandInterface):
            alchemy = self.alchemy
            bot = self
            loop = self.loop

            def register_herald_action(ci,
                                       event_name: str,
                                       coroutine: typing.Callable[[typing.Any], typing.Awaitable[typing.Dict]]) -> None:
                self.network_handlers[event_name] = coroutine

            def unregister_herald_action(ci, event_name: str):
                del self.network_handlers[event_name]

            async def call_herald_action(ci, destination: str, event_name: str, args: typing.Dict) -> typing.Dict:
                if self.network is None:
                    raise UnsupportedError("Herald is not enabled on this bot")
                request: rh.Request = rh.Request(handler=event_name, data=args)
                response: rh.Response = await self.network.request(destination=destination, request=request)
                if isinstance(response, rh.ResponseFailure):
                    if response.extra_info["type"] == "CommandError":
                        raise CommandError(response.extra_info["message"])
                    raise CommandError(f"Herald action call failed:\n"
                                       f"[p]{response}[/p]")
                elif isinstance(response, rh.ResponseSuccess):
                    return response.data
                else:
                    raise TypeError(f"Other Herald Link returned unknown response:\n"
                                    f"[p]{response}[/p]")

        return GenericInterface

    def _data_factory(self) -> typing.Type[CommandData]:
        raise NotImplementedError()

    def _init_network(self):
        """Create a :py:class:`royalherald.Link`, and run it as a :py:class:`asyncio.Task`."""
        if self.uninitialized_network_config is not None:
            self.network: rh.Link = rh.Link(self.uninitialized_network_config.master_uri,
                                            self.uninitialized_network_config.master_secret,
                                            self.interface_name,
                                            self._network_handler)
            log.debug(f"Running NetworkLink {self.network}")
            self.loop.create_task(self.network.run())

    async def _network_handler(self, message: typing.Union[rh.Request, rh.Broadcast]) -> rh.Response:
        try:
            network_handler = self.network_handlers[message.handler]
        except KeyError:
            log.warning(f"Missing network_handler for {message.handler}")
            return rh.ResponseFailure("no_handler", f"This bot is missing a network handler for {message.handler}.")
        else:
            log.debug(f"Using {network_handler} as handler for {message.handler}")
        if isinstance(message, rh.Request):
            try:
                response_data = await network_handler(self, **message.data)
                return rh.ResponseSuccess(data=response_data)
            except Exception as e:
                sentry_sdk.capture_exception(e)
                log.error(f"Exception {e} in {network_handler}")
                return rh.ResponseFailure("exception_in_handler",
                                          f"An exception was raised in {network_handler} for {message.handler}.",
                                          extra_info={
                                              "type": e.__class__.__name__,
                                              "message": str(e)
                                          })
        elif isinstance(message, rh.Broadcast):
            await network_handler(self, **message.data)

    def _init_database(self):
        """Create an :py:class:`royalnet.database.Alchemy` with the tables required by the packs. Then,
        find the chain that links the ``master_table`` to the ``identity_table``. """
        if self.uninitialized_database_config:
            log.info(f"Alchemy: enabled")
            required_tables = {self.uninitialized_database_config.master_table, self.uninitialized_database_config.identity_table}
            for command in self.uninitialized_commands:
                required_tables = required_tables.union(command.tables)
            log.debug(f"Required tables: {', '.join([item.__qualname__ for item in required_tables])}")
            self.alchemy = Alchemy(self.uninitialized_database_config.database_uri, required_tables)
            self.master_table = self.alchemy.__getattribute__(self.uninitialized_database_config.master_table.__name__)
            log.debug(f"Master table: {self.master_table.__qualname__}")
            self.identity_table = self.alchemy.__getattribute__(self.uninitialized_database_config.identity_table.__name__)
            log.debug(f"Identity table: {self.identity_table.__qualname__}")
            self.identity_column = self.identity_table.__getattribute__(self.identity_table,
                                                                        self.uninitialized_database_config.identity_column_name)
            log.debug(f"Identity column: {self.identity_column.__class__.__qualname__}")
            self.identity_chain = relationshiplinkchain(self.master_table, self.identity_table)
            log.debug(f"Identity chain: {' -> '.join([str(item) for item in self.identity_chain])}")
        else:
            log.info(f"Alchemy: disabled")
            self.alchemy = None
            self.master_table = None
            self.identity_table = None
            self.identity_column = None

    def _init_sentry(self):
        if self.uninitialized_sentry_dsn:
            # noinspection PyUnreachableCode
            if __debug__:
                release = "DEV"
            else:
                release = royalnet.version.semantic
            log.info(f"Sentry: enabled (Royalnet {release})")
            self.sentry = sentry_sdk.init(self.uninitialized_sentry_dsn,
                                          integrations=[AioHttpIntegration(),
                                                        SqlalchemyIntegration(),
                                                        LoggingIntegration(event_level=None)],
                                          release=release)
        else:
            log.info("Sentry: disabled")

    def _init_loop(self):
        if self.uninitialized_loop is None:
            self.loop = asyncio.get_event_loop()
        else:
            self.loop = self.uninitialized_loop

    def __init__(self, *,
                 network_config: typing.Optional[rh.Config] = None,
                 database_config: typing.Optional[DatabaseConfig] = None,
                 commands: typing.List[typing.Type[Command]] = None,
                 sentry_dsn: typing.Optional[str] = None,
                 loop: asyncio.AbstractEventLoop = None,
                 secrets_name: str = "__default__"):
        self.initialized = False
        self.uninitialized_network_config = network_config
        self.uninitialized_database_config = database_config
        self.uninitialized_commands = commands
        self.uninitialized_sentry_dsn = sentry_dsn
        self.uninitialized_loop = loop
        self.secrets_name = secrets_name

    def get_secret(self, username: str):
        return keyring.get_password(f"Royalnet/{self.secrets_name}", username)

    def set_secret(self, username: str, password: str):
        return keyring.set_password(f"Royalnet/{self.secrets_name}", username, password)

    def _initialize(self):
        if not self.initialized:
            self._init_sentry()
            self._init_loop()
            self._init_database()
            self._init_commands()
            self._init_network()
            self.initialized = True

    def run(self):
        """A blocking coroutine that should make the bot start listening to packs and requests."""
        raise NotImplementedError()

    def run_blocking(self, verbose=False):
        if verbose:
            core_logger = logging.root
            core_logger.setLevel(logging.DEBUG)
            stream_handler = logging.StreamHandler()
            stream_handler.formatter = logging.Formatter("{asctime}\t{name}\t{levelname}\t{message}", style="{")
            core_logger.addHandler(stream_handler)
            core_logger.debug("Logging setup complete.")
        self._initialize()
        self.loop.run_until_complete(self.run())



class CommandError(Exception):
    """Something went wrong during the execution of this command.

    Display an error message to the user, explaining what went wrong."""
    def __init__(self, message=""):
        self.message = message

    def __repr__(self):
        return f"{self.__class__.__qualname__}({repr(self.message)})"


class InvalidInputError(CommandError):
    """The command has received invalid input and cannot complete.

    Display an error message to the user, along with the correct syntax for the command."""


class UnsupportedError(CommandError):
    """A requested feature is not available on this interface.

    Display an error message to the user, telling them to use another interface."""


class KeyboardExpiredError(CommandError):
    """A special type of exception that can be raised in keyboard handlers to mark a specific keyboard as expired."""



import typing
import asyncio
import royalherald as rh
from .commanderrors import UnsupportedError
if typing.TYPE_CHECKING:
    from .command import Command
    from ..database import Alchemy
    from ..bots import GenericBot


class CommandInterface:
    name: str = NotImplemented
    prefix: str = NotImplemented
    alchemy: "Alchemy" = NotImplemented
    bot: "GenericBot" = NotImplemented
    loop: asyncio.AbstractEventLoop = NotImplemented

    def __init__(self):
        self.command: typing.Optional[Command] = None  # Will be bound after the command has been created

    def register_herald_action(self,
                               event_name: str,
                               coroutine: typing.Callable[[typing.Any], typing.Awaitable[typing.Dict]]):
        raise UnsupportedError(f"{self.register_herald_action.__name__} is not supported on this platform")

    def unregister_herald_action(self, event_name: str):
        raise UnsupportedError(f"{self.unregister_herald_action.__name__} is not supported on this platform")

    async def call_herald_action(self, destination: str, event_name: str, args: typing.Dict) -> typing.Dict:
        raise UnsupportedError(f"{self.call_herald_action.__name__} is not supported on this platform")

    def register_keyboard_key(self, key_name: str, callback: typing.Callable):
        raise UnsupportedError(f"{self.register_keyboard_key.__name__} is not supported on this platform")

    def unregister_keyboard_key(self, key_name: str):
        raise UnsupportedError(f"{self.unregister_keyboard_key.__name__} is not supported on this platform")



# Imports go here!
from .roll import RollCommand
from .dice import DiceCommand
from .dndactive import DndactiveCommand
from .dndinfo import DndinfoCommand
from .dndnew import DndnewCommand
from .dndedit import DndeditCommand
from .dndroll import DndrollCommand
from .dnditem import DnditemCommand
from .dndspell import DndspellCommand

# Enter the commands of your Pack here!
available_commands = [
    RollCommand,
    DiceCommand,
    DndactiveCommand,
    DndinfoCommand,
    DndnewCommand,
    DndeditCommand,
    DndrollCommand,
    DnditemCommand,
    DndspellCommand,
]

# Don't change this, it should automatically generate __all__
__all__ = [command.__class__.__qualname__ for command in available_commands]



import typing
import discord
from royalnet.commands import *
from royalnet.bots import DiscordBot


class PauseCommand(Command):
    name: str = "pause"

    description: str = "Mette in pausa o riprende la riproduzione della canzone attuale."

    syntax = "[ [guild] ]"

    @staticmethod
    async def _legacy_pause_handler(bot: DiscordBot, guild_name: typing.Optional[str]):
        # Find the matching guild
        if guild_name:
            guilds: typing.List[discord.Guild] = bot.client.find_guild_by_name(guild_name)
        else:
            guilds = bot.client.guilds
        if len(guilds) == 0:
            raise CommandError("No guilds with the specified name found.")
        if len(guilds) > 1:
            raise CommandError("Multiple guilds with the specified name found.")
        guild = list(bot.client.guilds)[0]
        # Set the currently playing source as ended
        voice_client: discord.VoiceClient = bot.client.find_voice_client_by_guild(guild)
        if not (voice_client.is_playing() or voice_client.is_paused()):
            raise CommandError("There is nothing to pause.")
        # Toggle pause
        resume = voice_client._player.is_paused()
        if resume:
            voice_client._player.resume()
        else:
            voice_client._player.pause()
        return {"resumed": resume}

    _event_name = "_legacy_pause"

    def __init__(self, interface: CommandInterface):
        super().__init__(interface)
        if interface.name == "discord":
            interface.register_herald_action(self._event_name, self._legacy_pause_handler)

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        guild_name, = args.match(r"(?:\[(.+)])?")
        response = await self.interface.call_herald_action("discord", self._event_name, {
                                                               "guild_name": guild_name
                                                           })
        if response["resumed"]:
            await data.reply(f"▶️ Riproduzione ripresa.")
        else:
            await data.reply(f"⏸ Riproduzione messa in pausa.")



import re
import random
from royalnet.commands import *
from ..tables import DndCharacter, DndActiveCharacter
from royalnet.utils import plusformat


class DndrollCommand(Command):
    name: str = "dndroll"

    description: str = "Roll dice as the active DnD character."

    aliases = ["dr", "dndr", "roll", "droll"]

    tables = {DndCharacter, DndActiveCharacter}

    _skill_names = {
        "str": "strength",
        "for": "strength",
        "dex": "dexterity",
        "des": "dexterity",
        "con": "constitution",
        "cos": "constitution",
        "inte": "intelligence",
        "wis": "wisdom",
        "sag": "wisdom",
        "cha": "charisma",
        "car": "charisma",

        "ststr": "strength_save",
        "stfor": "strength_save",
        "stdex": "dexterity_save",
        "stdes": "dexterity_save",
        "stcon": "constitution_save",
        "stcos": "constitution_save",
        "stint": "intelligence_save",
        "stwis": "wisdom_save",
        "stsag": "wisdom_save",
        "stcha": "charisma_save",
        "stcar": "charisma_save",

        "tsstr": "strength_save",
        "tsfor": "strength_save",
        "tsdex": "dexterity_save",
        "tsdes": "dexterity_save",
        "tscon": "constitution_save",
        "tscos": "constitution_save",
        "tsint": "intelligence_save",
        "tswis": "wisdom_save",
        "tssag": "wisdom_save",
        "tscha": "charisma_save",
        "tscar": "charisma_save",

        "acr": "acrobatics",
        "add": "animal_handling",
        "ani": "animal_handling",
        "arc": "arcana",
        "ath": "athletics",
        "dec": "deception",
        "ing": "deception",
        "his": "history",
        "sto": "history",
        "ins": "insight",
        "intu": "insight",
        "inti": "intimidation",
        "inv": "investigation",
        "med": "medicine",
        "nat": "nature",
        "perc": "perception",
        "perf": "performance",
        "pers": "persuasion",
        "rel": "religion",
        "sle": "sleight_of_hand",
        "soh": "sleight_of_hand",
        "rap": "sleight_of_hand",
        "ste": "stealth",
        "nas": "stealth",
        "sur": "survival",
        "sop": "sopravvivenza",
    }

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        author = await data.get_author(error_if_none=True)
        if author.dnd_active_character is None:
            raise CommandError("You don't have an active character.")
        char: DndCharacter = author.dnd_active_character.character

        first = args[0]
        second = args.optional(1)
        third = args.optional(2)

        advantage = False
        disadvantage = False
        extra_modifier = 0

        if third:
            try:
                extra_modifier = int(third)
            except ValueError:
                raise InvalidInputError("Invalid modifier value (third parameter).")
            if second.startswith("a") or second.startswith("v"):
                advantage = True
            elif second.startswith("d") or second.startswith("d"):
                disadvantage = True
            else:
                raise InvalidInputError("Invalid advantage string (second parameter).")

        elif second:
            try:
                extra_modifier = int(second)
            except ValueError:
                if second.startswith("a") or second.startswith("v"):
                    advantage = True
                elif second.startswith("d") or second.startswith("d"):
                    disadvantage = True
                else:
                    raise InvalidInputError("Invalid modifier value or advantage string (second parameter).")

        skill_short_name = first.lower()
        for root in self._skill_names:
            if skill_short_name.startswith(root):
                skill_name = self._skill_names[root]
                break
        else:
            raise CommandError("Invalid skill name (first parameter).")

        skill_modifier = char.__getattribute__(skill_name)
        modifier = skill_modifier + extra_modifier
        modifier_str = plusformat(modifier, empty_if_zero=True)

        if advantage:
            roll_a = random.randrange(1, 21)
            roll_b = random.randrange(1, 21)
            roll = max([roll_a, roll_b])
            total = roll + modifier
            await data.reply(f"🎲 2d20h1{modifier_str} = ({roll_a}|{roll_b}){modifier_str} = [b]{total}[/b]")
        elif disadvantage:
            roll_a = random.randrange(1, 21)
            roll_b = random.randrange(1, 21)
            roll = min([roll_a, roll_b])
            total = roll + modifier
            await data.reply(f"🎲 2d20l1{modifier_str} = ({roll_a}|{roll_b}){modifier_str} = [b]{total}[/b]")
        else:
            roll = random.randrange(1, 21)
            total = roll + modifier
            await data.reply(f"🎲 1d20{modifier_str} = {roll}{modifier_str} = [b]{total}[/b]")



import royalnet.version
import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

with open("requirements.txt", "r") as f:
    install_requires = f.readlines()

setuptools.setup(
    name="royalnet",
    version=royalnet.version.semantic,
    author="Stefano Pigozzi",
    author_email="ste.pigozzi@gmail.com",
    description="The great bot network of the User Games community",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/royal-games/royalnet",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Topic :: Internet",
        "Topic :: Database",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Video",
        "License :: OSI Approved :: MIT License"
    ],
    dependency_links=["https://github.com/Rapptz/discord.py/tarball/master"],
    include_package_data=True,
    zip_safe=False
)



"""Miscellaneous useful functions and classes."""

from .asyncify import asyncify
from .escaping import telegram_escape, discord_escape
from .safeformat import safeformat
from .classdictjanitor import cdj
from .sleepuntil import sleep_until
from .formatters import andformat, plusformat, fileformat, ytdldateformat, numberemojiformat, splitstring, ordinalformat

__all__ = [
    "asyncify",
    "safeformat",
    "cdj",
    "sleep_until",
    "plusformat",
    "andformat",
    "plusformat",
    "fileformat",
    "ytdldateformat",
    "numberemojiformat",
    "telegram_escape",
    "discord_escape",
    "splitstring",
    "ordinalformat",
]



import typing
import warnings
from .commanderrors import UnsupportedError
from .commandinterface import CommandInterface
from ..utils import asyncify


class CommandData:
    def __init__(self, interface: CommandInterface):
        self._interface: CommandInterface = interface
        if len(self._interface.command.tables) > 0:
            self.session = self._interface.alchemy.Session()
        else:
            self.session = None

    async def session_commit(self):
        """Commit the changes to the session."""
        await asyncify(self.session.commit)

    async def session_close(self):
        """Close the opened session.

        Remember to call this when the data is disposed of!"""
        if self.session:
            await asyncify(self.session.close)
            self.session = None

    async def reply(self, text: str) -> None:
        """Send a text message to the channel where the call was made.

        Parameters:
             text: The text to be sent, possibly formatted in the weird undescribed markup that I'm using."""
        raise UnsupportedError("'reply' is not supported on this platform")

    async def get_author(self, error_if_none: bool = False):
        """Try to find the identifier of the user that sent the message.
        That probably means, the database row identifying the user.

        Parameters:
            error_if_none: Raise an exception if this is True and the call has no author."""
        raise UnsupportedError("'get_author' is not supported on this platform")

    async def keyboard(self, text: str, keyboard: typing.Dict[str, typing.Callable]) -> None:
        """Send a keyboard having the keys of the dict as keys and calling the correspondent values on a press.

        The function should be passed the :py:class:`CommandData` instance as a argument."""
        warnings.warn("keyboard is deprecated, please avoid using it", category=DeprecationWarning)
        raise UnsupportedError("'keyboard' is not supported on this platform")

    async def delete_invoking(self, error_if_unavailable=False) -> None:
        """Delete the invoking message, if supported by the interface.

        The invoking message is the message send by the user that contains the command.

        Parameters:
            error_if_unavailable: if True, raise an exception if the message cannot been deleted."""
        if error_if_unavailable:
            raise UnsupportedError("'delete_invoking' is not supported on this platform")



import typing
import royalherald as rh


class RoyalnetRequestError(Exception):
    """An error was raised while handling the Royalnet request.

    This exception contains the :py:class:`royalherald.ResponseFailure` that was returned by the other Link."""
    def __init__(self, error: rh.ResponseFailure):
        self.error: rh.ResponseFailure = error

    @property
    def args(self):
        return f"{self.error.name}", f"{self.error.description}", f"{self.error.extra_info}"


class RoyalnetResponseError(Exception):
    """The :py:class:`royalherald.Response` that was received is invalid."""



from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import *
from ..utils import DndProficiencyType


class DndCharacter:
    __tablename__ = "dndcharacters"

    @declared_attr
    def character_id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def creator_id(self):
        return Column(Integer, ForeignKey("users.uid"))

    @declared_attr
    def creator(self):
        return relationship("User", foreign_keys=self.creator_id, backref="dndcharacters_created")

    @declared_attr
    def name(self):
        return Column(String, nullable=False)

    @declared_attr
    def strength_score(self):
        return Column(Integer, nullable=False)

    @property
    def strength(self):
        return (self.strength_score - 10) // 2

    @declared_attr
    def dexterity_score(self):
        return Column(Integer, nullable=False)

    @property
    def dexterity(self):
        return (self.dexterity_score - 10) // 2

    @declared_attr
    def constitution_score(self):
        return Column(Integer, nullable=False)

    @property
    def constitution(self):
        return (self.constitution_score - 10) // 2

    @declared_attr
    def intelligence_score(self):
        return Column(Integer, nullable=False)

    @property
    def intelligence(self):
        return (self.intelligence_score - 10) // 2

    @declared_attr
    def wisdom_score(self):
        return Column(Integer, nullable=False)

    @property
    def wisdom(self):
        return (self.wisdom_score - 10) // 2

    @declared_attr
    def charisma_score(self):
        return Column(Integer, nullable=False)

    @property
    def charisma(self):
        return (self.charisma_score - 10) // 2

    @declared_attr
    def level(self):
        return Column(Integer, nullable=False)

    @property
    def proficiency_bonus(self):
        return ((self.level - 1) // 4) + 2

    @declared_attr
    def current_hp(self):
        return Column(Integer, nullable=False)

    @declared_attr
    def max_hp(self):
        return Column(Integer, nullable=False)

    @declared_attr
    def armor_class(self):
        return Column(Integer, nullable=False)

    @declared_attr
    def strength_save_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def dexterity_save_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def constitution_save_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def intelligence_save_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def wisdom_save_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def charisma_save_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def acrobatics_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def animal_handling_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def arcana_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def athletics_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def deception_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def history_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def insight_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def intimidation_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def investigation_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def medicine_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def nature_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def perception_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def performance_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def persuasion_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def religion_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def sleight_of_hand_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def stealth_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @declared_attr
    def survival_proficiency(self):
        return Column(Enum(DndProficiencyType), nullable=False, default=DndProficiencyType.NONE)

    @property
    def strength_save(self):
        return self.strength + self.proficiency_bonus * self.strength_save_proficiency.value

    @property
    def dexterity_save(self):
        return self.dexterity + self.proficiency_bonus * self.dexterity_save_proficiency.value

    @property
    def constitution_save(self):
        return self.constitution + self.proficiency_bonus * self.constitution_save_proficiency.value

    @property
    def intelligence_save(self):
        return self.intelligence + self.proficiency_bonus * self.intelligence_save_proficiency.value

    @property
    def wisdom_save(self):
        return self.wisdom + self.proficiency_bonus * self.wisdom_save_proficiency.value

    @property
    def charisma_save(self):
        return self.charisma + self.proficiency_bonus * self.charisma_save_proficiency.value

    @property
    def acrobatics(self):
        return self.dexterity + self.proficiency_bonus * self.acrobatics_proficiency.value

    @property
    def animal_handling(self):
        return self.wisdom + self.proficiency_bonus * self.animal_handling_proficiency.value

    @property
    def arcana(self):
        return self.intelligence + self.proficiency_bonus * self.arcana_proficiency.value

    @property
    def athletics(self):
        return self.strength + self.proficiency_bonus * self.athletics_proficiency.value

    @property
    def deception(self):
        return self.charisma + self.proficiency_bonus * self.deception_proficiency.value

    @property
    def history(self):
        return self.intelligence + self.proficiency_bonus * self.history_proficiency.value

    @property
    def insight(self):
        return self.wisdom + self.proficiency_bonus * self.insight_proficiency.value

    @property
    def intimidation(self):
        return self.charisma + self.proficiency_bonus * self.intimidation_proficiency.value

    @property
    def investigation(self):
        return self.intelligence + self.proficiency_bonus * self.investigation_proficiency.value

    @property
    def medicine(self):
        return self.wisdom + self.proficiency_bonus * self.medicine_proficiency.value

    @property
    def nature(self):
        return self.intelligence + self.proficiency_bonus * self.nature_proficiency.value

    @property
    def perception(self):
        return self.wisdom + self.proficiency_bonus * self.perception_proficiency.value

    @property
    def performance(self):
        return self.charisma + self.proficiency_bonus * self.performance_proficiency.value

    @property
    def persuasion(self):
        return self.charisma + self.proficiency_bonus * self.persuasion_proficiency.value

    @property
    def religion(self):
        return self.intelligence + self.proficiency_bonus * self.religion_proficiency.value

    @property
    def sleight_of_hand(self):
        return self.dexterity + self.proficiency_bonus * self.sleight_of_hand_proficiency.value

    @property
    def stealth(self):
        return self.dexterity + self.proficiency_bonus * self.stealth_proficiency.value

    @property
    def survival(self):
        return self.wisdom + self.proficiency_bonus * self.survival_proficiency.value

    def __repr__(self):
        return f"<{self.__class__.__qualname__} {self.name}>"

    def __str__(self):
        return f"{self.name}"

    def character_sheet(self) -> str:
        columns = list(self.__class__.__table__.columns)
        column_names = [column.name for column in columns if (not column.primary_key and
                                                              not column.foreign_keys and
                                                              column.name != "name")]
        message = f"[b]{self.name}[/b]\n"
        for column_name in column_names:
            value = self.__getattribute__(column_name)
            message += f"{column_name} {value}\n"
        return message



import typing
import random
from royalnet.commands import *


class RollCommand(Command):
    name: str = "roll"

    description: str = "Roll a dice, from N to M (defaults to 1-100)."

    syntax = "[min] [max]"

    aliases = ["r", "random"]

    async def run(self, args: CommandArgs, data: CommandData) -> None:
        first: typing.Optional[str] = args.optional(0)
        second: typing.Optional[str] = args.optional(1)
        if second:
            minimum = int(first)
            maximum = int(second)
        elif first:
            minimum = 1
            maximum = int(first)
        else:
            minimum = 1
            maximum = 100
        result = random.randrange(minimum, maximum+1)
        await data.reply(f"🎲 Dice roll [{minimum}-{maximum}]: [b]{result}[/b]")
